﻿
Imports System.Collections.Generic

Module RS485_Module

    Friend SentPackets As Int32
    Friend ReceivedPackets As Int32

    ' -------------------------------------------------------------------- SLOT TO SERIAL
    Friend Sub RS485_SlotToSerial(ByVal Tdata As TransferData)
        If Tdata.Options.Contains("S") Then
            ' ----------------------------------------------- Accumulate command
            AccumulateCommand(Tdata)
        Else
            ' ----------------------------------------------- Execute command
            With Tdata
                Dim value As Int32 = SingleToInt32(Slots.ReadSlot_NoNan(.Slot))
                '
                Dim params(.NumBytes) As Byte
                '
                params(0) = .Address
                '
                IntegerToParamsArray(value, .NumBytes, params)
                '
                Dim response(5) As Byte

                If .Options.Contains("W") Then
                    WriteAndRead(.Device, 4, params, response)  ' Wait the ACTION command
                Else
                    WriteAndRead(.Device, 3, params, response)  ' Do it immediately
                End If
            End With
        End If
    End Sub

    ' -------------------------------------------------------------------- ACTION COMMAND
    Friend Sub RS485_ActionCommand()
        Dim params(-1) As Byte
        WriteAndRead(254, 5, params, Nothing)
    End Sub

    ' -------------------------------------------------------------------- SYNC COMANDS ACCUMULATOR
    Private Sub AccumulateCommand(ByVal Tdata As TransferData)
        TransferDataAccumulator.Add(Tdata)
    End Sub

    ' -------------------------------------------------------------------- SYNC COMMAND
    Friend Sub RS485_SyncCommand()
        If TransferDataAccumulator.Count < 1 Then Return
        ' ---------------------------------------------------------- test if they are all similar
        Dim ad As Byte = TransferDataAccumulator(0).Address
        Dim nb As Byte = TransferDataAccumulator(0).NumBytes
        For Each tdata As TransferData In TransferDataAccumulator
            If ad <> tdata.Address Or nb <> tdata.NumBytes Then
                Return
            End If
        Next
        ' ---------------------------------------------------------- create and fill the params byte array
        Dim params() As Byte
        ReDim params(2 + TransferDataAccumulator.Count * (nb + 1) - 1)
        params(0) = ad
        params(1) = nb
        '
        Dim index As Int32 = 2
        For Each tdata As TransferData In TransferDataAccumulator
            With tdata
                params(index) = .Device
                Dim value As Int32 = SingleToInt32(Slots.ReadSlot_NoNan(.Slot))
                IntegerToParamsArray(value, .NumBytes, params, index)
                index += (nb + 1)
            End With
        Next
        ' ----------------------------------------------------------- write the Sync-Write command
        WriteAndRead(254, &H83, params, Nothing)
        ' ----------------------------------------------------------- clear the TransferDataAccumulator list
        TransferDataAccumulator.Clear()
    End Sub

    ' -------------------------------------------------------------------- IMMEDIATE TO SERIAL
    Friend Sub RS485_ImmediateToSerial(ByVal Tdata As TransferData)
        With Tdata
            Dim params(.NumBytes) As Byte
            params(0) = .Address
            IntegerToParamsArray(.ImmediateValue, .NumBytes, params)
            Dim response(5) As Byte
            WriteAndRead(.Device, 3, params, response)
        End With
    End Sub

    ' -------------------------------------------------------------------- SET DEVICE ID
    Friend Sub RS485_SetDeviceIdentifier(ByVal ID_Enable As Int32, _
                                         ByVal ID_Address As Int32, _
                                         ByVal ID_Value As Int32)
        FastTimer_Stop()
        '
        RS485_BroadcastSingleByte(ID_Enable, 0)
        RS485_BroadcastSingleByte(ID_Address, ID_Value)
        RS485_BroadcastSingleByte(ID_Enable, 1)
        '
        Threading.Thread.Sleep(250)
        '
        If RS485_ReadSingleByte(ID_Value, ID_Address) <> ID_Value Then
            MessageBox.Show("ERROR - The device is not working" + vbCrLf + vbCrLf + _
                            " - Test the COM port and the BaudRate" + vbCrLf + _
                            " - Test the power supply and cables" + vbCrLf + _
                            " - Ensure a single device connected" + vbCrLf + _
                            " - Reprogram the device firmware" + vbCrLf + _
                            " - Reinstall the CH340 driver" + vbCrLf + _
                            " - Replace USB cable and connection module", _
                            "Message from Theremino-RS485", _
                            MessageBoxButtons.OK, MessageBoxIcon.Error)
        Else
            Beep()
        End If
        '
        FastTimer_Start()
    End Sub

    Friend Sub RS485_BroadcastSingleByte(ByVal address As Int32, _
                                         ByVal value As Int32)
        Dim params(1) As Byte
        params(0) = CByte(address)
        IntegerToParamsArray(value, 1, params)
        WriteAndRead(254, 3, params, Nothing)
    End Sub

    Friend Function RS485_ReadSingleByte(ByVal device As Int32, _
                                         ByVal address As Int32) As Byte
        Dim params(1) As Byte
        params(0) = CByte(address)
        params(1) = 1
        Dim response(5 + 1) As Byte
        WriteAndRead(CByte(device), 2, params, response)
        Return response(5)
    End Function

    ' -------------------------------------------------------------------- INSERT an INTEGER value converted to 1/2/3/4 bytes in params array
    Private Sub IntegerToParamsArray(ByVal value As Int32, _
                                     ByVal NumBytes As Int32, _
                                     ByRef params() As Byte, _
                                     Optional ByVal index As Int32 = 0)
        '
        ' .........................................................................
        '  TODO - Test with 3 and 4 bytes
        '  Dynamixel: The Size of data varies from 1 ~ 4 bytes. 
        '  For data larger than 2 bytes will be saved according to Little Endian.
        ' .........................................................................
        Select Case NumBytes
            Case 1
                Dim b() As Byte = BitConverter.GetBytes(value)
                params(index + 1) = b(0)
            Case 2
                If value < 0 Then value = -32768 - value
                Dim b() As Byte = BitConverter.GetBytes(value)
                params(index + 1) = b(0)
                params(index + 2) = b(1)
            Case 3
                Dim b() As Byte = BitConverter.GetBytes(value)
                params(index + 1) = b(0)
                params(index + 2) = b(1)
                params(index + 3) = b(2)
            Case 4
                Dim b() As Byte = BitConverter.GetBytes(value)
                params(index + 1) = b(0)
                params(index + 2) = b(1)
                params(index + 3) = b(2)
                params(index + 4) = b(3)
        End Select
    End Sub

    ' -------------------------------------------------------------------- SERIAL TO SLOT 
    Friend Sub RS485_SerialToSlot(ByVal Tdata As TransferData)
        With Tdata
            Dim params(1) As Byte
            params(0) = .Address
            params(1) = .NumBytes
            '
            Dim response(5 + .NumBytes) As Byte
            '
            WriteAndRead(.Device, 2, params, response)
            '
            Dim v As Single
            v = BufferToSingle(response, 5, .NumBytes)
            '
            Slots.WriteSlot(.Slot, v)
        End With
    End Sub

    ' -------------------------------------------------------------------- SERIAL TO BUFFER 
    Private Buffer(256) As Byte
    Friend Sub RS485_SerialToBuffer(ByVal Tdata As TransferData)
        With Tdata
            Dim params(1) As Byte
            params(0) = .Address
            params(1) = .NumBytes
            '
            Dim response(5 + .NumBytes) As Byte
            '
            WriteAndRead(.Device, 2, params, response)
            '
            For i As Int32 = 0 To .NumBytes - 1
                Buffer(.Address + i) = response(5 + i)
            Next
        End With
    End Sub

    ' -------------------------------------------------------------------- BUFFER TO SLOT 
    Friend Sub RS485_BufferToSlot(ByVal Tdata As TransferData)
        With Tdata
            Dim v As Single
            v = BufferToSingle(Buffer, .Address, .NumBytes)
            Slots.WriteSlot(.Slot, v)
        End With
    End Sub

    ' -------------------------------------------------------------------- CONVERT 1/2/3/4 bytes from BUFFER to SINGLE
    ' .........................................................................
    '  TODO - Test with 3 and 4 bytes
    '  Dynamixel: The Size of data varies from 1 ~ 4 bytes. 
    '  For data larger than 2 bytes will be saved according to Little Endian.
    ' .........................................................................
    Private Function BufferToSingle(ByVal buf() As Byte, _
                                    ByVal Index As Int32, _
                                    ByVal NumBytes As Int32) As Single
        Select Case NumBytes
            Case 1
                Return buf(Index)
            Case 2
                Dim n As Int32 = BitConverter.ToInt16(buf, Index)
                If n < 0 Then n = -32768 - n
                Return n
            Case 3
                Return buf(Index) + _
                       buf(Index + 1) * 256UI + _
                       buf(Index + 2) * 65536UI
            Case 4
                Return BitConverter.ToUInt32(buf, Index)
        End Select
    End Function


    ' -------------------------------------------------------------------- Broadcast_Baudrate
    Friend BaudRates(-1) As Int32
    Friend BaudRateRegisterIndex As Int32 = -1
    '
    Friend Sub BroadcastBaudrate(ByVal PortName As String, ByVal baudrate As Int32, ByVal ID_Enable As Int32)
        '
        Dim arrayindex As Int32 = Array.IndexOf(BaudRates, baudrate)
        If arrayindex < 0 Or arrayindex > 255 Then Return
        If BaudRateRegisterIndex < 0 Or BaudRateRegisterIndex > 255 Then Return
        ' ---------------------------------------------------------------- set baudrate combo items
        SetBaudrateComboItems()
        Form1.cmb_ComSpeed.Text = baudrate.ToString
        ' ---------------------------------------------------------------- send the message using all the valid baudrates 
        For i As Int32 = 0 To BaudRates.Length - 1
            '
            If i = arrayindex Then Continue For
            '
            COM_Open(PortName, BaudRates(i))
            If ComPort.IsOpen Then
                RS485_BroadcastSingleByte(ID_Enable, 0)
                RS485_BroadcastSingleByte(BaudRateRegisterIndex, arrayindex)
                RS485_BroadcastSingleByte(ID_Enable, 1)
            End If
        Next
    End Sub

    Friend Sub SetBaudrateComboItems()
        With Form1.cmb_ComSpeed
            Dim oldevents As Boolean = EventsAreEnabled
            EventsAreEnabled = False
            .Items.Clear()
            If BaudRates.Length > 0 Then
                For i As Int32 = 0 To BaudRates.Length - 1
                    .Items.Add(BaudRates(i).ToString)
                Next
            Else
                Dim b As Object() = {"1000000", "500000", _
                                     "256000", "250000", "128000", "115200", "74880", "57600", "56000", _
                                     "38400", "28800", "19200", "14400", "9600", "4800", "2400", "1200", "600"}
                .Items.AddRange(b)
            End If
            EventsAreEnabled = oldevents
        End With
    End Sub

    ' ===========================================================================
    '  WRITE AND READ
    ' ===========================================================================
    Private Sub WriteAndRead(ByVal PacketID As Byte, _
                             ByVal Instruction As Byte, _
                             ByVal params As Byte(), _
                             ByVal response As Byte())
        '
        If ComPort.IsOpen Then
            ' -------------------------------------------------------------- Clear in/out buffers
            ComPort.DiscardOutBuffer()
            ComPort.DiscardInBuffer()
            ' -------------------------------------------------------------- Packet is 255 + 255 + ID + LENGTH + INSTRUCTION + N(PARAMS) + CRC
            Dim packet(5 + params.Length) As Byte
            packet(0) = 255
            packet(1) = 255
            packet(2) = PacketID
            packet(3) = CByte(params.Length + 2)
            packet(4) = Instruction
            For i As Int32 = 0 To params.Length - 1
                packet(5 + i) = params(i)
            Next
            packet(packet.Length - 1) = Checksum(packet)
            ' -------------------------------------------------------------- Send packet to Serial Port
            Try
                ComPort.Write(packet, 0, packet.Length)
                SentPackets += 1
                ' ---------------------------------------------------------- wait response
                If PacketID <> 254 Then
                    GetResponse(response)
                    ' ------------------------------------------------------ test response
                    If TestPacketChecksum(response) Then
                        ReceivedPackets += 1
                    End If
                End If
            Catch
            End Try
        End If
    End Sub

    ' =======================================================================================================
    '  GET RESPONSE
    ' =======================================================================================================
    ' Someone says that there is a bug in the DataReceived Event (that doesn't fire all of the time).  
    ' =======================================================================================================
    '  CPU=30% FPS=180/350/210/420 (with 58+58/58/8+8/8 readings)
    'Private Sub GetResponse(ByRef response As Byte())
    '    For i As Integer = 0 To response.Length - 1
    '        Try
    '            response(i) = CByte(ComPort.ReadByte() And 255)
    '        Catch
    '        End Try
    '        If StopThreadFlag Then Return
    '    Next
    'End Sub

    '  CPU=1%  FPS=220/430/180/350 (with 58+58/58/8+8/8 readings)
    'Private Sub GetResponse(ByRef response As Byte())
    '    Do Until ComPort.BytesToRead >= response.Length
    '        If StopThreadFlag Then Return
    '        Threading.Thread.Sleep(1)
    '    Loop
    '    ComPort.Read(response, 0, response.Length)
    'End Sub

    ' --------------------------------------------------------------- THE BEST VERSION
    '  CPU=1% FPS=450/980/220/450 (with 58+58/58/8+8/8 readings)
    Private Sub GetResponse(ByRef response As Byte())
        COM_WaitDataReceived(response.Length)
        ComPort.Read(response, 0, response.Length)
    End Sub


    ' =======================================================================================================
    '  CHECKSUM
    ' =======================================================================================================
    Private Function TestPacketChecksum(ByVal packet() As Byte) As Boolean
        If packet(packet.Length - 1) = Checksum(packet) Then
            Return True
        Else
            Return False
        End If
    End Function

    Private Function Checksum(ByVal packet() As Byte) As Byte
        Dim ck As Int32
        For i As Int32 = 2 To packet.Length - 2
            ck = ck + packet(i)
        Next
        ck = ck And 255
        Return Not CByte(ck)
    End Function

End Module