﻿
Imports System.Collections.Generic
Imports System.Collections.Specialized

Module Module_RichTextBox
    Friend RTB_Lines() As String

    Friend Function ValidLine_ImmediateToSerial(ByVal sa() As String) As Boolean
        If Not IsNumeric(sa(0)) Then Return False
        If sa(1) <> "to" Then Return False
        If sa(2) <> "dxp1" Then Return False
        If Not IsNumeric(sa(3)) Then Return False
        If Not ValidDevice(sa(3)) Then Return False
        If Not DecodeAddress(sa(4), 4) Then Return False
        Return True
    End Function

    Friend Function ValidLine_SlotToSerial(ByVal sa() As String) As Boolean
        If Not IsNumeric(sa(1)) Then Return False
        If Not ValidSlotNumbers(sa(1)) Then Return False
        If sa(2) <> "to" Then Return False
        If sa(3) <> "dxp1" Then Return False
        If Not IsNumeric(sa(4)) Then Return False
        If Not ValidDevice(sa(4)) Then Return False
        If Not DecodeAddress(sa(5), 4) Then Return False
        If Not Valid_SlotToSerial_WithSync(sa) Then Return False
        Return True
    End Function

    Friend Function ValidLine_SerialToSlot(ByVal sa() As String) As Boolean
        If Not IsNumeric(sa(1)) Then Return False
        If Not ValidDevice(sa(1)) Then Return False
        If Not DecodeAddress(sa(2), 4) Then Return False
        If sa(3) <> "to" Then Return False
        If sa(4) <> "slot" Then Return False
        If Not IsNumeric(sa(5)) Then Return False
        If Not ValidSlotNumbers(sa(5)) Then Return False
        Return True
    End Function

    Friend Function ValidLine_Action(ByVal sa() As String) As Boolean
        If sa(0) <> "action" Then Return False
        If sa(1) <> "to" Then Return False
        If sa(2) <> "dxp1" Then Return False
        Return True
    End Function

    Friend Function ValidLine_Sync(ByVal sa() As String) As Boolean
        If sa(0) <> "sync" Then Return False
        If sa(1) <> "to" Then Return False
        If sa(2) <> "dxp1" Then Return False

        Return True
    End Function

    Friend Function ValidLine_SerialToBuffer(ByVal sa() As String) As Boolean
        If Not IsNumeric(sa(1)) Then Return False
        If Not ValidDevice(sa(1)) Then Return False
        If Not DecodeAddress(sa(2), 254) Then Return False
        If sa(3) <> "to" Then Return False
        If sa(4) <> "buffer" Then Return False
        Return True
    End Function
    Friend Function ValidLine_BufferToSlot(ByVal sa() As String) As Boolean
        If Not DecodeAddress(sa(1), 4) Then Return False
        If sa(2) <> "to" Then Return False
        If sa(3) <> "slot" Then Return False
        If Not IsNumeric(sa(4)) Then Return False
        If Not ValidSlotNumbers(sa(4)) Then Return False
        Return True
    End Function

    Friend Function ValidLine_SectionSelectionSlot(ByVal sa As String()) As Boolean
        If Not IsNumeric(sa(1)) Then Return False
        If Val(sa(1)) < 1 Then Return False
        If Val(sa(1)) > 999 Then Return False
        Return True
    End Function

    Private Function ValidDevice(ByVal s As String) As Boolean
        Dim n As Int32 = CInt(Val(s))
        If n > 0 Or n < 255 Then
            Return True
        Else
            Return False
        End If
    End Function

    Private Function ValidSlotNumbers(ByVal s1 As String) As Boolean
        If s1.Length > 3 Then Return False
        Dim v1 As Int32 = CInt(Val(s1))
        If v1 < 1 Then Return False
        If v1 > 999 Then Return False
        Return True
    End Function


    Friend Function ValidLine_BaudRates(ByVal sa() As String) As Boolean
        Dim sa2() As String = sa(1).Split(" "c)
        For i As Int32 = 0 To sa2.Length - 1
            If Not ValidBaudRate(sa2(i)) Then
                Return False
            End If
        Next
        Return True
    End Function
    Friend Function GetBaudRatesArray(ByVal sa() As String) As Int32()
        Dim sa2() As String = sa(1).Split(" "c)
        Dim ia(sa2.Length - 1) As Int32
        For i As Int32 = 0 To sa2.Length - 1
            ia(i) = StringToBaudRate(sa2(i))
        Next
        Return ia
    End Function
    Private Function ValidBaudRate(ByVal s As String) As Boolean
        Dim br As Int32 = StringToBaudRate(s)
        If br < 9600 Then Return False
        If br > 10000000 Then Return False
        Return True
    End Function
    Private Function StringToBaudRate(ByVal s As String) As Int32
        Dim br As Int32 = StringToInt32(s)
        If s.EndsWith("k") Then br *= 1000
        If s.EndsWith("m") Then br *= 1000000
        Return br
    End Function

    Friend Function ValidLine_BaudRateRegister(ByVal sa() As String) As Boolean
        If IsNumeric(sa(1)) Then
            If ValidRegister(sa(1)) Then
                Return True
            End If
        End If
        Return False
    End Function

    Private Function ValidRegister(ByVal s As String) As Boolean
        If Not IsNumeric(s) Then Return False
        If Val(s) < 0 Or Val(s) > 255 Then Return False
        Return True
    End Function

    Friend Function ValidLoopTimeMillisec(ByVal s As String) As Boolean
        If Val(s) < 0 Or Val(s) > 100 Then Return False
        Return True
    End Function

    Friend Function DecodeSlot(ByVal s As String) As Int32
        Return StringToInt32(s)
    End Function

    Friend Function DecodeDevice(ByVal s As String) As Byte
        Return StringToByte(s)
    End Function

    Friend AddressIndex As Byte
    Friend AddressBytes As Byte
    Friend Function DecodeAddress(ByVal s As String, ByVal MaxBytes As Int32) As Boolean
        Dim sa() As String = s.Split("-"c)
        If sa.Length <> 2 Then Return False
        AddressIndex = StringToByte(sa(0))
        AddressBytes = StringToByte(sa(1))
        If AddressBytes < 1 Then Return False
        If AddressBytes > MaxBytes Then Return False
        If CInt(AddressIndex) + AddressBytes > 254 Then Return False
        Return True
    End Function

    ' ALWAYS / SYNC / WAIT valid combinations
    Friend Function ValidOptions_SlotToSerial(ByVal s As String) As Boolean
        If s.Length > 3 Then Return False
        If Not "ASAWA".Contains(s.ToUpper) Then Return False
        Return True
    End Function


    ' -----------------------------------------------------------------------------------
    '  SlotToSerial lines with the SYNC option must have all the same address and length
    ' -----------------------------------------------------------------------------------
    Private SyncAddrIndex As Int32 = -1
    Private SyncAddrBytes As Int32
    Private Function Valid_SlotToSerial_WithSync(ByVal sa() As String) As Boolean
        If sa.Length > 6 Then
            If sa(6).Contains("s") Then
                DecodeAddress(sa(5), 4)
                If SyncAddrIndex < 0 Then
                    SyncAddrIndex = AddressIndex
                    SyncAddrBytes = AddressBytes
                Else
                    If SyncAddrIndex <> AddressIndex Then Return False
                    If SyncAddrBytes <> AddressBytes Then Return False
                End If
            End If
        End If
        Return True
    End Function
    Friend Sub SyncLineEncountered()
        SyncAddrIndex = -1
    End Sub

End Module

Public Class SyntaxRichTextBox
    Inherits System.Windows.Forms.RichTextBox

    Friend Sub InitProps()
        Me.AcceptsTab = True
        Me.DetectUrls = False
        Me.WordWrap = False
        Me.HideSelection = False
        ' The font-size determines the number of spaces for each TAB
        ' with font-size 15 each TAB is similar to 4 spaces
        Me.Font = New Font("Lucida Console", 15, FontStyle.Regular)
        Me.AutoWordSelection = True
        Me.AutoWordSelection = False
        Me.ZoomFactor = 0.8
    End Sub

    Private Sub SyntaxRichTextBox_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SelectionChanged
        ShowLinesCount()
    End Sub

    Friend Sub ShowLinesCount()
        RTB_Lines = Me.Lines
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        Dim nCol As Int32 = GetColumnFromCharIndex(SelectionStart)
        Form1.Label_Lines.Text = "Lines: " + RTB_Lines.Length.ToString + _
                                 "  Line: " + (nline + 1).ToString + _
                                 "  Col: " + nCol.ToString
    End Sub

    ' -------------------------------- 15 uS (independent from n-lines) -------
    Private Function GetColumnFromCharIndex(ByVal charindex As Int32) As Int32
        Dim n As Int32 = 0
        Dim nline As Int32 = Me.GetLineFromCharIndex(charindex)
        Dim nchars As Int32 = charindex - GetFirstCharIndexFromLine(nline)
        For i As Int32 = 0 To nchars - 1
            If RTB_Lines(nline)(i) = CChar(vbTab) Then
                n += 4 - (n Mod 4)
            Else
                n += 1
            End If
        Next
        Return n
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  COLOR LINES AND REPLACE KEYWORDS
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    ' ===============================================================================================
    '  COLORS
    ' ===============================================================================================
    Private Color_Text As Color = Color.FromArgb(0, 0, 220)
    Private Color_KeywordText As Color = Color.FromArgb(100, 0, 0)
    Private Color_CommentText As Color = Color.FromArgb(30, 30, 40)   ' 0,50,50
    Private Color_CommentBack As Color = Color.FromArgb(240, 245, 220)  '240, 250, 220

    Private Color_ActiveBack As Color = Color.FromArgb(255, 255, 160)
    Private Color_WarningBack As Color = Color.FromArgb(255, 220, 120)
    Private Color_NotFoundBack As Color = Color.FromArgb(255, 200, 170)
    Private Color_ErrorBack As Color = Color.FromArgb(255, 180, 150)

    Friend Function IsErrorOrWarningColor(ByVal c As Color) As Boolean
        If c.R > c.G + 10 Then
            Return True
        Else
            Return False
        End If
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  TOKENIZER
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Friend Function Tokenize(ByVal l As String) As String()
        l = ReplaceMultipleSpacesAndTabAndTrim(l).ToLower
        RemoveComments(l)
        Dim sa() As String = l.Split(" "c)
        Select Case sa(0)
            Case "baudrates"
                ReDim Preserve sa(1)
                sa(1) = l.Remove(0, 10)
        End Select
        Return sa
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  COLOR LINES
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Private EnabledArea As Boolean

    Friend Sub ColorAllLines()
        Form1.RTB.PaintEventEnabled = False
        Form1.RTB.Visible = True
        ' ----------------------------------------------------------
        Dim sw As Stopwatch = New Stopwatch
        sw.Start()
        ' ----------------------------------------------------------
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTB_Lines.Length - 1
            ' ------------------------------------------------------ AutomationLanguageColors
            ColorSingleLine_AndReplaceKeywords(i)
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ----------------------------------------------------------
        Form1.RTB.Enabled = True
        Form1.RTB.PaintEventEnabled = True
    End Sub

    Private Sub ColorLines(ByVal startline As Int32, ByVal endline As Int32)
        RTB_Lines = Me.Lines
        For i As Int32 = startline To endline
            ColorSingleLine(i)
        Next
    End Sub

    Private Sub ColorSingleLine(ByVal i As Int32)
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- save position 
        Dim OldCursorPosition As Integer = SelectionStart
        ' ---------------------------------------------------------- set the whole line to default colors
        SelectLine(i)
        Me.SelectionColor = Color_Text
        Me.SelectionBackColor = Color_ActiveBack
        ' ---------------------------------------------------------- AutomationLanguageColors
        ColorSingleLine_AndReplaceKeywords(i)
        ' ---------------------------------------------------------- restore
        SelectionStart = OldCursorPosition
        SelectionLength = 0
        ' ---------------------------------------------------------- 
        EventsAreEnabled = old
    End Sub

    Friend Sub ColorSelectedLine()
        RTB_Lines = Me.Lines
        ColorSingleLine(GetLineFromCharIndex(SelectionStart))
    End Sub


    Private LineWithErrors As Boolean

    Private Sub ColorSingleLine_AndReplaceKeywords(ByVal nline As Int32)
        If nline < 0 Or nline > RTB_Lines.Length - 1 Then Return
        ' --------------------------------------------------------------- 
        Dim line As String = RTB_Lines(nline)
        ' --------------------------------------------------------------- 
        If line <> "" Then
            ' ----------------------------------------------------------- color active back
            SelectLine(nline)
            Me.SelectionColor = Color_Text
            Me.SelectionBackColor = Color_ActiveBack
            ' ----------------------------------------------------------- comment position
            Dim CommentPosition As Int32 = RemoveComments(line)
            If CommentPosition = 0 Then
                MarkCommentLine(nline, 0)
                Return
            End If
            '
            LineWithErrors = False
            ' ----------------------------------------------------------- tokenize to array of keywords
            Dim sa() As String = Tokenize(line.Trim.ToLowerInvariant)
            ' ----------------------------------------------------------- 
            Select Case sa.Length
                Case 1
                    Select Case sa(0)
                        Case "end"
                            EnabledArea = False
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 2
                    Select Case sa(0)
                        Case "baudrates"
                            If Not ValidLine_BaudRates(sa) Then
                                MarkErrorLine(nline)
                            End If
                        Case "baudrateregister"
                            If Not ValidLine_BaudRateRegister(sa) Then
                                MarkErrorLine(nline)
                            End If
                        Case "looptimemillisec"
                            If Not ValidLoopTimeMillisec(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "sectionselectorslot"
                            If Not ValidLine_SectionSelectionSlot(sa) Then
                                MarkErrorLine(nline)
                            End If
                        Case "section"
                            If Not IsNumeric(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 3
                    Select Case sa(0)
                        Case "action"
                            If Not ValidLine_Action(sa) Then
                                MarkErrorLine(nline)
                            End If
                        Case "sync"
                            If ValidLine_Sync(sa) Then
                                SyncLineEncountered()
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 5
                    Select Case sa(0)
                        Case "dxp1"
                            If Not ValidLine_SerialToBuffer(sa) Then MarkErrorLine(nline)
                        Case "buffer"
                            If Not ValidLine_BufferToSlot(sa) Then MarkErrorLine(nline)
                        Case Else
                            If Not ValidLine_ImmediateToSerial(sa) Then MarkErrorLine(nline)
                    End Select
                Case 6
                    Select Case sa(0)
                        Case "slot"
                            If Not ValidLine_SlotToSerial(sa) Then MarkErrorLine(nline)
                        Case "dxp1"
                            If Not ValidLine_SerialToSlot(sa) Then MarkErrorLine(nline)
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 7
                    Select Case sa(0)
                        Case "slot"
                            If Not ValidLine_SlotToSerial(sa) Then MarkErrorLine(nline)
                            If Not ValidOptions_SlotToSerial(sa(6)) Then MarkErrorLine(nline)
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case Else
                    MarkErrorLine(nline)
            End Select
            ' ----------------------------------------------------------- color the comment section
            If CommentPosition > -1 Then
                MarkCommentLine(nline, CommentPosition)
            End If
            ' ----------------------------------------------------------- replace keywords
            'If Not LineWithErrors Then
            ReplaceKeywords(nline, RTB_Lines(nline), CommentPosition)
            'End If
        End If
    End Sub

    Private Sub MarkErrorLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_ErrorBack
        LineWithErrors = True
    End Sub
    Private Sub MarkNotFoundLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_NotFoundBack
    End Sub
    Private Sub MarkWarningLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_WarningBack
    End Sub
    Private Sub MarkCommentLine(ByVal i As Int32, Optional ByVal FirstCommentedChar As Int32 = -1)
        SelectLine(i)
        If FirstCommentedChar > 0 Then
            SelectionLength -= FirstCommentedChar ' lenght must be changed before start
            SelectionStart += FirstCommentedChar  ' otherwise at the end of the file the lenght is shortened 
        End If
        Me.SelectionColor = Color_CommentText
        Me.SelectionBackColor = Color_CommentBack
    End Sub

    Private Sub MarkDisabledLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionColor = Color_CommentText
        Me.SelectionBackColor = Color_CommentBack
    End Sub


    ' ===============================================================================================
    '  REPLACE KEYWORDS
    ' ===============================================================================================
    Private Keywords() As String = {"Slot", "DXP1", "Buffer", "To", "Action", "Sync", "BaudRates", "BaudRateRegister", _
                                    "LoopTimeMillisec", "SectionSelectorSlot", "Section", "End"}

    Friend Sub ReplaceKeywords(ByVal lineNumber As Int32, _
                               ByRef line As String, _
                               ByVal commentPosition As Int32)
        Dim LineStart As Int32
        If lineNumber >= 0 Then LineStart = GetFirstCharIndexFromLine(lineNumber)
        ' --------------------------------------------------------------- SPECIAL LINES
        Dim line2 As String = line.Trim.ToLower
        '
        If line2.StartsWith("beep") Then
            ReplaceKeywords(New String() {"Beep"}, lineNumber, LineStart, line, commentPosition)
        Else
            ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)                  ' REPLACE ALL KEYWORDS
            'ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS
        End If
    End Sub

    Private Sub ReplaceKeywords(ByVal Keywords As String(), _
                                ByVal lineNumber As Int32, _
                                ByVal lineStart As Int32, _
                                ByRef line As String, _
                                ByVal commentPosition As Int32, _
                                Optional ByVal UseTextColor As Boolean = False)

        If commentPosition < 0 Then commentPosition = 9999
        Dim position As Int32
        For Each kw As String In Keywords
            position = 0
            Do
                position = line.IndexOf(kw, position, StringComparison.InvariantCultureIgnoreCase)
                If position >= 0 Then
                    If IsCompleteWord(line, position, kw.Length) And position < commentPosition Then
                        If Not InsideQuotes(position, line) Then
                            If lineNumber >= 0 Then
                                Me.Select(lineStart + position, kw.Length)
                                Me.SelectionColor = If(UseTextColor, Color_Text, Color_KeywordText)
                                Me.Select(lineStart + position, kw.Length) ' <-- redo selection after SelectionColor 
                                If kw <> Me.SelectedText Then
                                    Me.SelectedText = kw
                                End If
                            Else
                                Mid(line, position + 1, kw.Length) = kw
                            End If
                        End If
                    End If
                    position += kw.Length
                End If
            Loop Until position < 0
        Next
    End Sub

    Private Function IsCompleteWord(ByVal l As String, ByVal position As Int32, ByVal length As Int32) As Boolean
        ' ------------------------------------------------------------ test left side
        If position > 1 AndAlso _
           IsLetterOrSpaceOrNumber(l(position - 1)) Then Return False
        ' ------------------------------------------------------------- test right side
        If position + length < l.Length AndAlso _
           IsLetterOrSpaceOrNumber(l(position + length)) Then Return False
        ' -------------------------------------------------------------
        Return True
    End Function

    Private Function IsLetterOrSpaceOrNumber(ByVal c As Char) As Boolean
        Return (c >= "a" And c <= "z") OrElse (c >= "A" And c <= "Z") OrElse (c >= "0" And c <= "9") OrElse c = "_"
    End Function

    Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
        InsideQuotes = False
        If index >= text.Length Then Return False
        For Counter As Int32 = 0 To index - 1
            If text(Counter) = """"c Then
                InsideQuotes = Not InsideQuotes
            End If
        Next
    End Function

    Friend Sub DeleteFinalEmptyLines()
        RTB_Lines = Me.Lines
        For i As Int32 = RTB_Lines.Length - 1 To 0 Step -1
            If RTB_Lines(i).Trim = "" Then
                ReDim Preserve RTB_Lines(RTB_Lines.Length - 2)
            Else
                Exit For
            End If
        Next
        Me.Lines = RTB_Lines
        Me.Text += vbCrLf
    End Sub


    ' ===============================================================================================
    '  MIXED FUNCTIONS
    ' ===============================================================================================
    Friend Sub ScrollToCenterLine(ByVal line As Int32)
        Dim VisibleLinesCount As Int32 = GetLastVisibleLine() - GetFirstVisibleLine()
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        PaintEventEnabled = False
        Enabled = False
        Dim delta As Int32 = VisibleLinesCount \ 2 - 1
        ScrollToLine(line + delta)
        ScrollToLine(line - delta)
        Enabled = True
        PaintEventEnabled = True
        EventsAreEnabled = old
    End Sub

    Friend Sub SelectCompleteWord_Short()
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub

    Friend Sub SelectCompleteWord_Long()
        Dim start, length As Int32
        FindWordPointers(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub


    ' ===============================================================================================
    '  ENABLE DISABLE PAINT and DISABLE ZOOM with CTRL WHEEL
    ' ===============================================================================================
    Private Const MK_CONTROL As Int32 = &H8
    Private Const WM_MOUSEWHEEL As Int32 = &H20A
    Friend PaintEventEnabled As Boolean = True
    <DebuggerStepThrough()> _
    Protected Overrides Sub WndProc(ByRef m As Message)
        Const WM_PAINT As Int32 = &HF
        If m.Msg = WM_PAINT AndAlso Not PaintEventEnabled Then
            m.Result = IntPtr.Zero
        ElseIf m.Msg = WM_MOUSEWHEEL Then
            If (m.WParam.ToInt32 And MK_CONTROL) = MK_CONTROL Then
                m.WParam = IntPtr.Zero
                m.Result = IntPtr.Zero
            Else
                MyBase.WndProc(m)
            End If
        Else
            MyBase.WndProc(m)
        End If
    End Sub

    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '   EVENTS
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    ' ===============================================================================================
    '  DOUBLE CLICK (select without the final space and including the underscore)
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.DoubleClick
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        SelectionStart = start
        SelectionLength = length
    End Sub

    ' ===============================================================================================
    '   KEY DOWN / KEY UP
    ' ===============================================================================================
    Private OverwriteMode As Boolean = False
    Private LastKeyPressed As Keys

    '  UNDO - Undo entries are saved only after important text changes
    Private UndoAndHistoryTextReady As Boolean = False

    Private Sub SyntaxRichTextBox_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        '
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete
            Case Keys.Left, Keys.Right, Keys.Up, Keys.Down, Keys.PageUp, Keys.PageDown, Keys.Home, Keys.End
            Case Else
                UndoAndHistoryTextReady = True
        End Select
        '
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete
                If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
                TextMustBeIndented = True
            Case Keys.Tab
                If SelectedText.Length > 0 Then
                    IndentSelectedLines(If(e.Shift, -1, 1))
                    e.Handled = True
                    e.SuppressKeyPress = True
                End If
            Case Keys.Insert
                OverwriteMode_Toggle()
        End Select
        ' -------------------------------------------------- 
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Up, Keys.Down
                TrimRightSpacesFromLine(Me.GetLineFromCharIndex(Me.SelectionStart))
        End Select
        ' -------------------------------------------------- 
        LastKeyPressed = e.KeyCode
        RTB_Lines = Me.Lines()
        ' -------------------------------------------------- suppress RTB sounds
        RichTextBox_SuppressSounds(e)
    End Sub

    Private Sub SyntaxRichTextBox_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        '
        ' -------------------------------------------------- THIS ENABLES MULTIPAGE SELECTION WITH SHIFT
        If e.KeyCode = Keys.ShiftKey Then Return
        ' -------------------------------------------------- 
        RTB_Lines = Me.Lines()
        ' -------------------------------------------------- show the Overwrite Caret
        OverwriteMode_ShowCaret(Me)
    End Sub


    ' ==============================================================
    '  RICH TEXT BOX - SUPPRESS SOUNDS (called from KeyDown)
    ' ==============================================================
    Friend Sub RichTextBox_SuppressSounds(ByRef e As KeyEventArgs)
        If Me.GetLineFromCharIndex(Me.SelectionStart) = 0 AndAlso _
                                            (e.KeyData = Keys.Up Or e.KeyData = Keys.PageUp) _
        OrElse Me.GetLineFromCharIndex(Me.SelectionStart) = Me.GetLineFromCharIndex(Me.TextLength) AndAlso _
                                            (e.KeyData = Keys.Down Or e.KeyData = Keys.PageDown) _
        OrElse Me.SelectionStart = Me.TextLength AndAlso e.KeyData = Keys.Right _
        OrElse Me.SelectionStart = 0 AndAlso _
                (e.KeyData = Keys.Left Or (e.KeyData = Keys.Back And Me.SelectionLength = 0)) Then
            e.Handled = True
        End If
    End Sub

    ' ======================================================================================================
    '  OVERWRITE MODE CARET
    ' ======================================================================================================
    Private Declare Function CreateCaret Lib "user32" (ByVal hWnd As IntPtr, _
                                                      ByVal hBitmap As IntPtr, _
                                                      ByVal nWidth As Int32, _
                                                      ByVal nHeight As Int32) As Int32

    Private Declare Function ShowCaret Lib "user32" (ByVal hWnd As IntPtr) As Int32

    Private CaretOverwriteMode As Boolean

    Friend Sub OverwriteMode_Toggle()
        CaretOverwriteMode = Not CaretOverwriteMode
        If Not OverwriteMode Then
            CreateCaret(Me.Handle, IntPtr.Zero, 1, CInt((Me.Font.Height - 1) * Me.ZoomFactor))
            ShowCaret(Me.Handle)
        End If
    End Sub

    Friend Sub OverwriteMode_ShowCaret(ByRef rtb As SyntaxRichTextBox)
        If OverwriteMode Then
            CreateCaret(rtb.Handle, IntPtr.Zero, CInt(12 * rtb.ZoomFactor), CInt((rtb.Font.Height - 1) * rtb.ZoomFactor))
            ShowCaret(rtb.Handle)
        End If
    End Sub


    ' ===============================================================================================
    '   MOUSE DOWN
    ' ===============================================================================================
    Private OldCursorLine As Int32 = -1

    Private MouseDownSelectionIndex As Int32

    Private Sub SyntaxRichTextBox_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        '
        MouseDownSelectionIndex = Me.GetCharIndexFromPosition(e.Location)
        '
        If e.Button = MouseButtons.Right And Me.SelectionLength = 0 Then
            Me.Select(MouseDownSelectionIndex, 0)
        End If
        '
        OverwriteMode_ShowCaret(Me)
    End Sub


    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers(ByVal text As String, _
                                ByVal selectionIndex As Int32, _
                                ByRef start As Int32, _
                                ByRef lenght As Int32)
        Const separators As String = " " + vbCr + vbLf + vbTab
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    '  This version find shortest words - not including characters like  " , / () ...
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers_ShortWords(ByVal text As String, _
                                            ByVal selectionIndex As Int32, _
                                            ByRef start As Int32, _
                                            ByRef lenght As Int32)
        Const separators As String = " "";:,/\|()'" + vbCr + vbLf + vbTab
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    Friend Sub ScrollToLine(ByVal nline As Int32)
        If RTB_Lines.Length = 0 Then Return
        If nline < 0 Then nline = 0
        If nline > RTB_Lines.Length - 1 Then nline = RTB_Lines.Length - 1
        Me.Select(GetFirstCharIndexFromLine(nline), 1)
        Me.ScrollToCaret()
    End Sub

    ' ===============================================================================================
    '   ON TEXT CHANGED
    ' ===============================================================================================
    Friend TextMustBeIndented As Boolean = False
    Friend TextModified As Boolean = False
    Private LastEditedLine As Int32 = -1
    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        If Not PaintEventEnabled Then Return
        ' ------------------------------------------------------- 
        RTB_Lines = Me.Lines
        ' ------------------------------------------------------- 
        Dim line As Int32 = GetLineFromCharIndex(SelectionStart)
        ' ------------------------------------------------------- stop Paint
        PaintEventEnabled = False
        ' No "Enabled = false" here
        ' otherwise the key autorepeat stops and the scroll bar flashes
        ' ------------------------------------------------------- process the line
        ColorSingleLine(line)
        ' ------------------------------------------------------- eventually process also the precedent line
        If line <> LastEditedLine Then
            ColorSingleLine(line - 1)
        End If
        ' ------------------------------------------------------- restart Paint
        PaintEventEnabled = True
        ' ------------------------------------------------------- 
        TextModified = True
        TextMustBeIndented = True
        ' ------------------------------------------------------- LastEditedLine <= 0 for undo deleting all the text
        If line <> LastEditedLine Or LastEditedLine <= 0 Then
            UndoAndHistory_AddEntry()
            LastEditedLine = line
        End If
        ' -------------------------------------------------------
        MyBase.OnTextChanged(e)
    End Sub

    ' ===============================================================================================
    '  GoToNextOccurrence / GoToDeclaration
    ' ===============================================================================================
    Friend Sub GoToNextOccurrence()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTB_Lines.Length - 1 Then Beep() : testLine = 0
            If testLine = startLine Then Exit For
            l = RTB_Lines(testLine).ToLower.Trim.Replace("""", "")
            'l = l.Replace("'", "")
            l = l.TrimEnd
            If l.Contains(s) Then
                If IsWholeWord(l, l.IndexOf(s), s) Then
                    Me.ScrollToCenterLine(testLine)
                    Dim n As Int32 = RTB_Lines(testLine).ToLower.IndexOf(s)
                    Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                    History_AddEntry()
                    Me.Focus()
                    Exit For
                End If
            End If
        Next
    End Sub

    Friend Sub GoToDeclaration()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTB_Lines.Length - 1 Then Beep() : testLine = 0
            If testLine = startLine Then Exit For
            l = RTB_Lines(testLine).ToLower.Trim.Replace("""", "")
            'l = l.Replace("'", "")
            l = l.TrimEnd
            RemoveComments(l)
            If l.Contains(s) Then
                If IsWholeWord(l, l.IndexOf(s), s) Then
                    Me.ScrollToCenterLine(testLine)
                    Dim n As Int32 = RTB_Lines(testLine).ToLower.IndexOf(s)
                    Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                    History_AddEntry()
                    Me.Focus()
                    Exit For
                End If
            End If
        Next
    End Sub


    ' ===============================================================================================
    '  INDENT-UNINDENT
    ' ===============================================================================================
    Friend Sub IndentSelectedLines(ByVal increment As Int32)
        If RTB_Lines.Length > 0 Then
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim ntabs As Int32 = CountInitialTabs_NotAllowingSpaces(lineText, 4) + increment
                If ntabs >= 0 Then
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    Me.Select(start, CountInitialChars(lineText))
                    Me.SelectedText = StrDup(ntabs, vbTab)
                End If
            Next
            ' ----------------------------------------------------------------
            SelectLines(FirstSelectedLine, LastSelectedLine)
            ' ----------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            UndoAndHistory_AddEntry()
        End If
        Me.Focus()
    End Sub

    ' ===============================================================================================
    '  COMMENT UN-COMMENT
    ' ===============================================================================================
    Friend Sub CommentSelectedLines(ByVal comment As Boolean)
        ' ------------------------------------------------------------ 
        RTB_Lines = Me.Lines
        If RTB_Lines.Length > 0 Then
            ' ----------------------------------------------------------------
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim CommentColumn As Int32 = -1
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch >= 0 Then
                    If CommentColumn < 0 Then CommentColumn = firstch
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)

                    ' CORRECTION FOR RS485 and similar apps
                    CommentColumn = 0

                    If comment Then
                        Me.Select(start + CommentColumn, 0)
                        Me.SelectedText = "'"
                    Else
                        Me.Select(start + firstch, 1)
                        If Me.SelectedText = "'" Then
                            Me.SelectedText = ""
                        End If
                    End If
                End If
            Next
            ' ------------------------------------------------------------
            ColorLines(FirstSelectedLine, LastSelectedLine)
            ' ------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            UndoAndHistory_AddEntry()
            ' ------------------------------------------------------------ 
            Me.SelectLines(FirstSelectedLine, LastSelectedLine)
            ' ------------------------------------------------------------ fire a text changed event
            MyBase.OnTextChanged(Nothing)
        End If
        Me.Focus()
    End Sub

    Private Function CountInitialTabs(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    Private Function CountInitialTabs_NotAllowingSpaces(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : Return 9999           ' Spaces are not allowed
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function


    ' ===============================================================================================
    '  CUT COPY and PASTE
    ' ===============================================================================================
    Friend Sub ExecCut()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
        End If
    End Sub
    Friend Sub ExecCopy()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, Me.SelectedText.Replace(vbLf, vbCrLf))
        Clipboard.SetDataObject(data, True)
    End Sub
    Friend Sub ExecPaste()
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            ' ------------------------------------------------------------ 
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ extract only the text from clipbbard
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            Me.SelectedText = Clipboard.GetText
            Dim endline As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            ' ------------------------------------------------------------ color lines
            ColorLines(startline, endline)
            ' ------------------------------------------------------------
            RTB_Lines = Me.Lines
            Me.SelectLines(startline, endline)
            Me.SelectionLength = 0
            Me.SelectionStart = Me.GetLastCharIndexFromLine(endline)
            ' ------------------------------------------------------------
            UndoAndHistory_AddEntry()
            Me.Enabled = True
            PaintEventEnabled = True
            Me.Focus()
            ' ------------------------------------------------------------ fire a text changed event
            MyBase.OnTextChanged(Nothing)
        End If
    End Sub
    Friend Sub ExecDelete()
        If Me.SelectionLength > 0 Then
            ' ------------------------------------------------------------ 
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ 
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            Me.SelectedText = ""
            Dim endline As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            ' ------------------------------------------------------------ color lines
            ColorLines(startline, endline)
            ' ------------------------------------------------------------
            RTB_Lines = Me.Lines
            Me.SelectLines(startline, endline)
            Me.SelectionLength = 0
            Me.SelectionStart = Me.GetLastCharIndexFromLine(endline)
            ' ------------------------------------------------------------
            UndoAndHistory_AddEntry()
            Me.Enabled = True
            PaintEventEnabled = True
            Me.Focus()
        End If
    End Sub


    ' ===============================================================================================
    '  FUNCTIONS
    ' ===============================================================================================
    Friend Function RemoveComments(ByRef s As String) As Int32
        Dim i As Int32 = s.IndexOf("'")
        If i > 0 Then s = s.Remove(i).TrimEnd
        If i = 0 Then s = ""
        Return i
    End Function
    Friend Sub SelectLine(ByVal line As Int32)
        If line >= 0 AndAlso line < RTB_Lines.Length AndAlso RTB_Lines.Length > 0 Then
            Dim lineStart As Int32 = GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = RTB_Lines(line).Length
            Me.Select(lineStart, lineLength)
        Else
            Me.Select(0, 0)
        End If
    End Sub
    Friend Sub SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        Dim FirstChar As Int32 = Me.GetFirstCharIndexFromLine(FirstLine)
        Dim LastChar As Int32 = Me.GetFirstCharIndexFromLine(LastLine) + RTB_Lines(LastLine).Length - 1
        Me.Select(FirstChar, LastChar - FirstChar)
    End Sub
    Friend Function GetFirstSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart)
    End Function
    Private Function GetLastSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
    End Function
    Friend Function GetFirstVisibleLine() As Int32
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    End Function
    Friend Function GetLastVisibleLine() As Int32
        Dim p As Point = New Point(1, Me.ClientSize.Height)
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(p))
    End Function
    Private Function CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function
    Friend Function GetLastCharIndexFromLine(ByVal nline As Int32) As Int32
        Return Me.GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length
    End Function
    Private Sub TrimRightSpacesFromLine(ByVal i As Int32)
        If i < 0 Or i >= RTB_Lines.Length Then Return
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        If RTB_Lines(i).Trim = "" Then
            While RTB_Lines(i).EndsWith(" ")
                RTB_Lines(i) = RTB_Lines(i).Remove(RTB_Lines(i).Length - 1, 1)
            End While
        Else
            RTB_Lines(i) = RTB_Lines(i).TrimEnd
        End If
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        If endpos > startpos Then
            Me.Select(startpos, endpos - startpos)
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
        End If
    End Sub

    ' ===============================================================================================================
    '   RTB - SAVE and RESTORE selection and scroll ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Friend Structure ScrollData
        Dim OldSelectionStart As Int32
        Dim OldSelectionLength As Int32
        Dim OldSelectionFirstChar As Int32
        Dim OldSelectionLastChar As Int32
    End Structure

    Friend Sub SaveSelectionAndScroll(ByRef sd As ScrollData)
        sd.OldSelectionStart = Me.SelectionStart
        sd.OldSelectionLength = Me.SelectionLength
        sd.OldSelectionFirstChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        sd.OldSelectionLastChar = Me.GetCharIndexFromPosition(New Point(1, Me.ClientSize.Height))
    End Sub

    Friend Sub RestoreSelectionAndScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        ' --------------------------------------------------------------
        If sd.OldSelectionStart >= sd.OldSelectionFirstChar Then
            Me.Select(sd.OldSelectionStart, sd.OldSelectionLength)
        End If
    End Sub

    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '   UNDO and HISTORY (history is unimplemented)
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Friend Sub UndoAndHistory_AddEntry()
        ' ----------------------------------------------------------------
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        Undo_AddEntry()
        'History_AddEntry()
        ' ----------------------------------------------------------------
        UndoAndHistoryTextReady = False
    End Sub

    ' ===============================================================================================
    '  PRIVATE UNDO FUNCTIONS
    ' ===============================================================================================
    Private Structure UndoEntry
        Dim ActionName As String
        Dim RTF_Text As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private UndoBuffer As List(Of UndoEntry) = New List(Of UndoEntry)
    Private Undo_Index As Int32
    Friend Sub Undo_Clear()
        UndoBuffer.Clear()
        Undo_Index = 0
        UndoAndHistory_AddEntry()
        TextModified = False
    End Sub
    Private Sub Undo_AddEntry()
        Dim newentry As UndoEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        newentry.RTF_Text = Me.Rtf
        ' ----------------------------------------------------------------
        While UndoBuffer.Count > Undo_Index
            UndoBuffer.RemoveAt(UndoBuffer.Count - 1)
        End While
        ' ----------------------------------------------------------------
        UndoBuffer.Add(newentry)
        Undo_Index += 1
        'Debug.Print(UndoBuffer.Count.ToString)
        ' ----------------------------------------------------------------
        TextModified = True
        TextMustBeIndented = True
    End Sub

    ' ===============================================================================================
    '  UNDO and REDO commands
    ' ===============================================================================================
    Friend Sub ExecUndo()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        Undo_Index -= 1
        If Undo_Index < 1 Then Undo_Index = 1
        Dim undo As UndoEntry
        undo = UndoBuffer(Undo_Index - 1)
        Me.Rtf = undo.RTF_Text
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0)
        'Me.Select(undo.SelectionStart, undo.SelectionLength)
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- mark as not-modified
        If Undo_Index < 2 Then TextModified = False
        ' ----------------------------------------------------------------
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------------- fire a text changed event
        MyBase.OnTextChanged(Nothing)
    End Sub
    Friend Sub ExecRedo()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        If Undo_Index > UndoBuffer.Count - 1 Then Undo_Index = UndoBuffer.Count - 1
        Dim undo As UndoEntry = UndoBuffer(Undo_Index)
        Me.Rtf = undo.RTF_Text
        Undo_Index += 1
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0)
        'Me.Select(undo.SelectionStart, undo.SelectionLength)
        ' ---------------------------------------------------------------- 
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- mark as modified
        TextModified = True
        TextMustBeIndented = True
        ' ---------------------------------------------------------------- reinit user variables
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------------- fire a text changed event
        MyBase.OnTextChanged(Nothing)
    End Sub
    Friend Function Editor_CanUndo() As Boolean
        Return Undo_Index >= 2
    End Function
    Friend Function Editor_CanRedo() As Boolean
        Return Undo_Index <= UndoBuffer.Count - 1
    End Function

    ' ===============================================================================================
    '   PRIVATE HISTORY FUNCTIONS
    ' ===============================================================================================
    Private Structure HistoryEntry
        Dim ActionName As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private HistoryBuffer As List(Of HistoryEntry) = New List(Of HistoryEntry)
    Private History_Index As Int32
    Friend Sub History_Clear()
        HistoryBuffer.Clear()
        History_Index = 0
        History_AddEntry()
    End Sub
    Friend Sub History_AddEntry()
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        Dim newentry As HistoryEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        ' ----------------------------------------------------------------
        'While HistoryBuffer.Count > History_Index
        '    HistoryBuffer.RemoveAt(HistoryBuffer.Count - 1)
        'End While
        ' ----------------------------------------------------------------
        HistoryBuffer.Add(newentry)
        History_Index += 1
    End Sub

    ' ===============================================================================================
    '  HISTORY commands
    ' ===============================================================================================
    Friend Sub History_Back()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- back
        History_Index -= 1
        If History_Index < 1 Then History_Index = 1
        Dim hist As HistoryEntry
        hist = HistoryBuffer(History_Index - 1)
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        'Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- 
        ShowLinesCount()
    End Sub
    Friend Sub History_Forward()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- forward
        If History_Index > HistoryBuffer.Count - 1 Then History_Index = HistoryBuffer.Count - 1
        Dim hist As HistoryEntry = HistoryBuffer(History_Index)
        History_Index += 1
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        'Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ----------------------------------------------------------------
        ShowLinesCount()
    End Sub
    Friend Function Editor_CanHistoryBack() As Boolean
        Return History_Index >= 2
    End Function
    Friend Function Editor_CanHistoryForward() As Boolean
        Return History_Index <= HistoryBuffer.Count - 1
    End Function

    ' ===============================================================================================
    '   LOST FOCUS
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
        If UndoAndHistoryTextReady Then
            UndoAndHistory_AddEntry()
        End If
    End Sub

End Class





