﻿
Imports System.Net
Imports System.Collections.Generic

Module TransferFunctions

    Friend LoopTimeMillisec As Int32 = 1
    Friend SectionSelectorSlot As Int32 = -1
    Friend ActualSection As Int32
    Friend ActualSectionSlotValue As Int32

    Private TransferData_List As List(Of TransferData) = New List(Of TransferData)
    Friend TransferDataAccumulator As List(Of TransferData) = New List(Of TransferData)

    Friend Enum TransferTypes
        SerialToSlot
        SlotToSerial
        ImmediateToSerial
        SerialToBuffer
        BufferToSlot
        Action
        Sync
    End Enum

    Friend Class TransferData
        Public TransferType As TransferTypes
        Public Slot As Int32
        Public Device As Byte
        Public Address As Byte
        Public NumBytes As Byte
        Public OldValue As Int32
        Public Options As String
        Public ImmediateValue As Int32
        Public Executed As Boolean = False
        Public Sub New(ByVal TransferType As TransferTypes, ByVal Slot As Int32, _
                                                            ByVal Device As Byte, _
                                                            ByVal Address As Byte, _
                                                            ByVal NumBytes As Byte, _
                                                            Optional ByVal Options As String = "", _
                                                            Optional ByVal ImmediateValue As Int32 = 0)
            Me.TransferType = TransferType
            Me.Slot = Slot
            Me.Device = Device
            Me.Address = Address
            Me.NumBytes = NumBytes
            Me.OldValue = Int32.MinValue
            Me.Options = Options
            Me.ImmediateValue = ImmediateValue
        End Sub
    End Class

    ' ===========================================================================================================
    '   REINIT TRANSFER LISTS
    ' ===========================================================================================================
    Friend Sub ReinitTransferLists(ByVal la() As String)
        Dim sa As String()
        ' ------------------------------------------------------------------- Read SectionSelectorSlot and looptimemillisec before 
        SectionSelectorSlot = -1 ' disabled by default
        ReDim BaudRates(-1)
        BaudRateRegisterIndex = -1
        ActualSection = 0
        '
        For Each l As String In la
            ' --------------------------------------------------------------- Tokenize line
            sa = Form1.RTB.Tokenize(l)
            ' --------------------------------------------------------------- Decode line
            Select Case sa.Length
                ' ----------------------------------------------------------- Commands
                Case 2
                    Select Case sa(0)
                        Case "baudrates"
                            If ValidLine_BaudRates(sa) Then
                                BaudRates = GetBaudRatesArray(sa)
                            End If
                        Case "baudrateregister"
                            If ValidLine_BaudRateRegister(sa) Then
                                BaudRateRegisterIndex = StringToInt32(sa(1))
                            End If
                        Case "looptimemillisec"
                            If ValidLoopTimeMillisec(sa(1)) Then
                                LoopTimeMillisec = StringToInt32(sa(1))
                            End If
                        Case "sectionselectorslot"
                            If ValidLine_SectionSelectionSlot(sa) Then
                                SectionSelectorSlot = StringToInt32(sa(1))
                            End If
                    End Select
            End Select
        Next
        '
        ' ------------------------------------------------------------------- Read SERIAL commands
        TransferData_List.Clear()
        TransferDataAccumulator.Clear()
        For Each l As String In la
            l = ReplaceMultipleSpacesAndTabAndTrim(l).ToLower
            Form1.RTB.RemoveComments(l)
            sa = l.Split(" "c)
            ' --------------------------------------------------------------- Test section
            If SectionSelectorSlot > 0 Then
                If sa.Length = 2 AndAlso sa(0) = "section" Then
                    ActualSection = CInt(Val(sa(1)))
                End If
                If ActualSection <> ActualSectionSlotValue And ActualSection <> 0 Then
                    Continue For
                End If
            End If
            ' --------------------------------------------------------------- Decode line
            Select Case sa.Length
                ' ----------------------------------------------------------- END
                Case 1
                    Select Case sa(0)
                        Case "end"
                            Exit For
                    End Select
                Case 3
                    ' ------------------------------------------------------- Action Command
                    Select Case sa(0)
                        Case "action"
                            If ValidLine_Action(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.Action, _
                                                                       0, _
                                                                       0, _
                                                                       0, _
                                                                       0))
                            End If
                    End Select
                    ' ------------------------------------------------------- Sync Command
                    Select Case sa(0)
                        Case "sync"
                            If ValidLine_Sync(sa) Then
                                TransferData_List.Add(New TransferData(TransferTypes.Sync, _
                                                                         0, _
                                                                         0, _
                                                                         0, _
                                                                         0))
                                SyncLineEncountered()
                            End If
                    End Select
                Case 5
                    ' ------------------------------------------------------- SerialToBuffer / BufferToSlot
                    Select Case sa(0)
                        Case "dxp1"
                            If ValidLine_SerialToBuffer(sa) Then
                                DecodeAddress(sa(2), 254)
                                TransferData_List.Add(New TransferData(TransferTypes.SerialToBuffer, _
                                                                       0, _
                                                                       DecodeDevice(sa(1)), _
                                                                       AddressIndex, _
                                                                       AddressBytes))
                            End If
                        Case "buffer"
                            If ValidLine_BufferToSlot(sa) Then
                                DecodeAddress(sa(1), 4)
                                TransferData_List.Add(New TransferData(TransferTypes.BufferToSlot, _
                                                                       DecodeSlot(sa(4)), _
                                                                       1, _
                                                                       AddressIndex, _
                                                                       AddressBytes))
                            End If
                        Case Else
                            If ValidLine_ImmediateToSerial(sa) Then
                                DecodeAddress(sa(4), 4)
                                TransferData_List.Add(New TransferData(TransferTypes.ImmediateToSerial, _
                                                                       -1, _
                                                                       DecodeDevice(sa(3)), _
                                                                       AddressIndex, _
                                                                       AddressBytes, _
                                                                       "", _
                                                                        StringToInt32(sa(0))))
                            End If
                    End Select
                Case 6
                    ' ------------------------------------------------------- SlotToSerial / SerialToSlot
                    Select Case sa(0)
                        Case "slot"
                            If ValidLine_SlotToSerial(sa) Then
                                DecodeAddress(sa(5), 4)
                                TransferData_List.Add(New TransferData(TransferTypes.SlotToSerial, _
                                                                       DecodeSlot(sa(1)), _
                                                                       DecodeDevice(sa(4)), _
                                                                       AddressIndex, _
                                                                       AddressBytes))
                            End If
                        Case "dxp1"
                            If ValidLine_SerialToSlot(sa) Then
                                DecodeAddress(sa(2), 4)
                                TransferData_List.Add(New TransferData(TransferTypes.SerialToSlot, _
                                                                       DecodeSlot(sa(5)), _
                                                                       DecodeDevice(sa(1)), _
                                                                       AddressIndex, _
                                                                       AddressBytes))
                            End If
                    End Select
                Case 7
                    ' ------------------------------------------------------- SlotToSerial with options
                    Select Case sa(0)
                        Case "slot"
                            If ValidLine_SlotToSerial(sa) Then
                                If ValidOptions_SlotToSerial(sa(6)) Then
                                    DecodeAddress(sa(5), 4)
                                    TransferData_List.Add(New TransferData(TransferTypes.SlotToSerial, _
                                                                           DecodeSlot(sa(1)), _
                                                                           DecodeDevice(sa(4)), _
                                                                           AddressIndex, _
                                                                           AddressBytes, _
                                                                           sa(6).ToUpper))
                                End If
                            End If
                    End Select
            End Select
        Next
    End Sub

    Friend Sub UpdateSectionSelector()
        If SectionSelectorSlot > 0 Then
            Dim n As Int32 = CInt(Slots.ReadSlot(SectionSelectorSlot))
            If n <> ActualSectionSlotValue Then
                ActualSectionSlotValue = n
                RTB_Lines = Form1.RTB.Lines
                ReinitTransferLists(RTB_Lines)
            End If
        End If
    End Sub



    ' ===========================================================================================================
    '   READ WRITE REGISTERS
    ' ===========================================================================================================
    Friend Sub ReadWriteRegisters()
        If Not ComPort.IsOpen Then Return
        Try
            For Each tdata As TransferData In TransferData_List
                Select Case tdata.TransferType
                    Case TransferTypes.SlotToSerial
                        '
                        If tdata.Options.Contains("A") Then
                            ' ---------------------------------------------------- always
                            RS485_SlotToSerial(tdata)
                        Else
                            ' ---------------------------------------------------- only if changed
                            Dim v As Int32
                            v = SingleToInt32(Slots.ReadSlot_NoNan(tdata.Slot))
                            If v <> tdata.OldValue Then
                                tdata.OldValue = v
                                RS485_SlotToSerial(tdata)
                            End If
                        End If
                        '
                    Case TransferTypes.Action
                        RS485_ActionCommand()
                        '
                    Case TransferTypes.Sync
                        RS485_SyncCommand()
                        '
                    Case TransferTypes.ImmediateToSerial
                        If Not tdata.Executed Then
                            RS485_ImmediateToSerial(tdata)
                            tdata.Executed = True
                        End If
                        '
                    Case TransferTypes.SerialToSlot
                        RS485_SerialToSlot(tdata)
                        '
                    Case TransferTypes.SerialToBuffer
                        RS485_SerialToBuffer(tdata)
                        '
                    Case TransferTypes.BufferToSlot
                        RS485_BufferToSlot(tdata)
                        '
                End Select
            Next
            ' --------------------------------------------------------------------------------------- reduce CPU charge
            If ReceivedPackets = 0 Then
                Threading.Thread.Sleep(1)
            End If
            ' --------------------------------------------------------------------------------------- Update counters
            ExchangedFrames += 1
        Catch
            'Beep()
        End Try
    End Sub


    ' ===========================================================================================================
    '   COMMUNICATION STATISTICS
    ' ===========================================================================================================
    Friend Stat_Working As Double
    Friend Stat_FPS As Double
    Private ExchangedFrames As Int32

    Friend Sub Statistics_Update()
        ' ---------------------------------------------------------------- Working or Inactive
        If SentPackets = 0 Then
            Stat_Working -= 1
            If Stat_Working < 0 Then Stat_Working = 0
        Else
            Stat_Working += 1
            If Stat_Working > 15 Then Stat_Working = 15
        End If
        ' ---------------------------------------------------------------- measure the true milliseconds
        Static SW As Stopwatch = New Stopwatch
        Dim millisec As Double = SW.Elapsed.TotalMilliseconds
        If SW.IsRunning And millisec = 0 Then Return
        SW.Reset()
        SW.Start()
        ' ---------------------------------------------------------------- update Stat_FPS 
        If millisec > 0 Then
            SmoothValue_Pow_Adaptive(Stat_FPS, ExchangedFrames * 1000 / millisec, 10)
        End If
        ' ---------------------------------------------------------------- reset all values
        SentPackets = 0
        ReceivedPackets = 0
        ExchangedFrames = 0
    End Sub


    ' ===========================================================================================================
    '   FAST TIMER
    ' ===========================================================================================================
    Private FastTimerThread As Threading.Thread
    Friend StopThreadFlag As Boolean

    Friend Sub FastTimer_Start()
        FastTimer_Stop()
        ' --------------------------------------------------------- 
        SchedulerMaxPrecision()
        ' --------------------------------------------------------- start the Thread 
        FastTimerThread = New Threading.Thread(AddressOf FastThread_Execute)
        ' --------------------------------------------------------- make the thread as background thread.
        FastTimerThread.IsBackground = True
        ' --------------------------------------------------------- set the Priority of the thread.
        FastTimerThread.Priority = Threading.ThreadPriority.AboveNormal
        ' --------------------------------------------------------- start the thread
        StopThreadFlag = False
        FastTimerThread.Start()
    End Sub

    Friend Sub FastTimer_Stop()
        If FastTimerThread Is Nothing Then Return
        If Not FastTimerThread.IsAlive Then Return
        ' --------------------------------------------- try to join the thread
        COM_WaitDataReceived_Stop()
        StopThreadFlag = True
        FastTimerThread.Join(10)
        ' --------------------------------------------- try catch because sometimes the thread is already closed
        Try
            While FastTimerThread.IsAlive
                Threading.Thread.Sleep(10)
                Application.DoEvents()
            End While
            ' ----------------------------------------- eventually abort
            If FastTimerThread.IsAlive Then
                MessageBox.Show("ExecutionThread not closed correctly.", _
                                "Message from Theremino-485", _
                                MessageBoxButtons.OK, MessageBoxIcon.Warning)
                FastTimerThread.Abort()
            End If
        Catch
            Try
                FastTimerThread.Abort()
            Catch
            End Try
        End Try
        ' ---------------------------------------------
        FastTimerThread = Nothing
        ' ---------------------------------------------
        SchedulerDefaultPrecision()
    End Sub

    Private Sub FastThread_Execute()
        Do
            ReadWriteRegisters()
            Threading.Thread.Sleep(LoopTimeMillisec)
        Loop Until StopThreadFlag
    End Sub

    ' =========================================================================
    '  WINDOWS SCHEDULER PRECISION
    ' ========================================================================= 
    Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Sub SchedulerMaxPrecision()
        If OperatingSystemIsWindows Then
            timeBeginPeriod(1)
        End If
    End Sub
    Private Sub SchedulerDefaultPrecision()
        If OperatingSystemIsWindows Then
            timeEndPeriod(1)
        End If
    End Sub

End Module