﻿Module Module_Utils


    Friend Function IsWholeWord(ByVal Text As String, _
                        ByVal Position As Int32, _
                        ByVal Word As String) As Boolean
        '
        Static Word_Chars As String = "ABCDEFGHIJKLMNOPQRSTUVWYXZabcdefghijklmnopqrstuvwyxz0123456789_@"
        '
        If (Position <= 0 OrElse Not Word_Chars.Contains(Text(Position - 1))) AndAlso _
           (Position + Len(Word) >= Text.Length OrElse Not Word_Chars.Contains(Text(Position + Len(Word)))) Then
            Return True
        Else
            Return False
        End If
    End Function

    ' ----------------------------------------------------------------------------------------------
    '  Extract from "text" the complete word pointed by the "selectionIndex" (short words version)
    ' ----------------------------------------------------------------------------------------------
    Friend Function GetCompleteWord_Short(ByVal text As String, _
                                          ByVal selectionIndex As Int32) As String
        Dim start, length As Int32
        FindWordPointers_ShortWords(text, selectionIndex, start, length)
        Return text.Substring(start, length)
    End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    '  This version find shortest words - not including characters like  " , / () ...
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers_ShortWords(ByVal text As String, _
                                           ByVal selectionIndex As Int32, _
                                           ByRef start As Int32, _
                                           ByRef lenght As Int32)
        '
        Const separators As String = " "";:,/\|()'" + vbCr + vbLf + vbTab
        '
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If i1 - 1 < text.Length AndAlso InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ======================================================================================================
    '  OPEN HELP FILE
    ' ======================================================================================================
    Friend HelpFile_PathAndName As String

    Friend Sub OpenHelpFile()
        ' ------------------------------------------------------------------------------------------------
        If TryToOpenFile(HelpFile_PathAndName) Then Return
        If TryToOpenHelp(Application.StartupPath + "\Docs") Then Return
        ' ------------------------------------------------------------------------------------------------
        Form_Input.Text = "Message from Theremino-485"
        Form_Input.Prompt = "Help file not found" + vbCrLf + vbCrLf
        Form_Input.Prompt += "Help files are searched in the ""Docs"" folder, in the following languages order:" + vbCrLf
        Form_Input.Prompt += "ENG, FRA, SPA, ITA, POR, JPN, CHI" + vbCrLf + vbCrLf
        Form_Input.Prompt += "Press the 'OPEN DOCS FOLDER' button and copy your preferred Help file." + vbCrLf + vbCrLf
        Form_Input.Prompt += "Otherwise press the 'SELECT' button and select your Help file location."
        Form_Input.txt_Input.Visible = False
        Form_Input.chk_OpenDocsFolder.Text = "OPEN DOCS FOLDER"
        Form_Input.chk_Select.Text = "SELECT"
        Dim OldFormInputLocation As Point = Form_Input.Location
        Dim OldFormInputSize As Size = Form_Input.Size
        Dim OldCancelButtonSize As Size = Form_Input.chk_OpenDocsFolder.Size
        Form_Input.Left = Form1.Left + 120
        Form_Input.Top = Form1.Top + 110
        Form_Input.Width = 580
        Form_Input.Height = 270
        Form_Input.chk_OpenDocsFolder.Size = New Size(180, Form_Input.chk_OpenDocsFolder.Height)
        Form_Input.ShowDialog(Form1)
        Select Case Form_Input.SelectedInput
            Case "Select"
                LocateHelpFile()
            Case "OpenFolder"
                Dim s As String = Application.StartupPath + "\Docs"
                My.Computer.FileSystem.CreateDirectory(s)
                Process.Start(s)
        End Select
        Form_Input.Location = OldFormInputLocation
        Form_Input.Size = OldFormInputSize
        Form_Input.chk_OpenDocsFolder.Size = OldCancelButtonSize
    End Sub

    Private Function TryToOpenHelp(ByVal path As String) As Boolean
        path += "\Theremino_Automation_Help"
        If TryToOpenFile(path + "_ENG.pdf") Then Return True
        If TryToOpenFile(path + "_ESP.pdf") Then Return True
        If TryToOpenFile(path + "_FRA.pdf") Then Return True
        If TryToOpenFile(path + "_ITA.pdf") Then Return True
        If TryToOpenFile(path + "_POR.pdf") Then Return True
        If TryToOpenFile(path + "_GER.pdf") Then Return True
        If TryToOpenFile(path + "_CHI.pdf") Then Return True
        If TryToOpenFile(path + "_JPN.pdf") Then Return True
        Return False
    End Function
    Private Function TryToOpenFile(ByVal file As String) As Boolean
        If My.Computer.FileSystem.FileExists(file) Then
            Process.Start(file)
            Return True
        Else
            Return False
        End If
    End Function

    Friend Sub LocateHelpFile()
        Dim ofd As OpenFileDialog = New OpenFileDialog()
        'ofd.InitialDirectory = PrepareSaveLoadFolder()
        ' ---------------------------------- left part of filename not visible
        ofd.FileName = IO.Path.GetFileName(HelpFile_PathAndName)
        ofd.DefaultExt = ".pdf"
        ' ------------------------ filename completely visible but no extension
        'ofd.FileName = IO.Path.GetFileNameWithoutExtension(FilePathAndName)
        'ofd.DefaultExt = IO.Path.GetExtension(FilePathAndName)
        ' ------------------------ filename completely visible and better dialog
        ofd.ShowHelp = True
        ' ------------------------------------------------------------
        'ofd.RestoreDirectory = True
        ofd.AutoUpgradeEnabled = True
        ofd.Title = "Select help file"
        ofd.ValidateNames = False
        ofd.Multiselect = True
        ofd.Filter = "All files (*.*)|*.*"
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            HelpFile_PathAndName = ofd.FileName
            TryToOpenFile(HelpFile_PathAndName)
            Save_INI()
        End If
    End Sub


    ' ==============================================================================================================
    '   ToolTips
    ' ==============================================================================================================
    Private WithEvents ttp1 As ToolTip = New ToolTip
    Friend Sub ToolTips_Init()
        ttp1.ShowAlways = True
        ttp1.UseAnimation = True
        ttp1.UseFading = True
        ttp1.IsBalloon = True
        ttp1.AutoPopDelay = 32700
        ttp1.InitialDelay = 500
        ttp1.ReshowDelay = 500
        ' --------------------------------------------------------------- Add here Controls and ToolTip-Text
        ttp1.SetToolTip(Form1.chk_Comment, "Comment the selected text")
        ttp1.SetToolTip(Form1.chk_Uncomment, "Uncomment the selected text")
        ttp1.SetToolTip(Form1.chk_Undo, "Undo the last changes")
        ttp1.SetToolTip(Form1.chk_Redo, "Redo the last changes")
        ttp1.SetToolTip(Form1.chk_HistoryBack, "History back")
        ttp1.SetToolTip(Form1.chk_HistoryForward, "History forward")
        ttp1.SetToolTip(Form1.chk_GoToNextOccurrence, "Find next occurrence")
        ttp1.SetToolTip(Form1.chk_GotoDeclaration, "Find next occurrence (active words only)")
        ttp1.SetToolTip(Form1.chk_Options, "Open the option window." + vbCrLf + "(set device identifiers)")
        ttp1.SetToolTip(Form1.chk_Help, "Open the HELP document." + vbCrLf + "Press the right mouse button to select the Help file")

        'ToolStripButton_Validate.ToolTipText = Msg_Validate1
        'ToolStripButton_BeepOnErrors.ToolTipText = Msg_BeepOnErrors1
        'ToolStripButton_Lock.ToolTipText = Msg_Lock1 & vbCrLf & Msg_Lock2
        'ToolStripButton_Disconnect.ToolTipText = Msg_Disconnect1
    End Sub

    ' --------------------- this corrects the XP bug that causes tooltip disappearing for ever 
    'Private Sub controls_MouseEnter(ByVal sender As Object, _
    '                               ByVal e As System.EventArgs) Handles txt_CommSpeed.MouseEnter, _
    '                                                                    txt_MinChange.MouseEnter, _
    '                                                                    txt_CapArea.MouseEnter
    '    ttp1.Active = False
    '    ttp1.Active = True
    'End Sub


    ' =======================================================================================================
    '  ONLY Positive Integer NUMERIC COMBO BOX
    ' -------------------------------------------------------------------------------------------------------
    '  Call from KeyDown event with: OnlyNumericComboBox(sender, e)
    ' =======================================================================================================
    Friend Sub OnlyNumericComboBox(ByVal sender As Object, ByVal e As KeyEventArgs)
        ' ------------------------------------------------------ Allow navigation keyboard arrows
        Select Case e.KeyCode
            Case Keys.Up, Keys.Down, Keys.Left, Keys.Right, Keys.PageUp, Keys.PageDown, Keys.Delete
                e.SuppressKeyPress = False
                Return
        End Select
        ' ------------------------------------------------------ Block non-number characters
        Dim currentKey As Char = Chr(e.KeyCode)
        If Not My.Computer.Keyboard.CtrlKeyDown And _
           Not Char.IsControl(currentKey) And _
           Not Char.IsDigit(currentKey) And _
           Not e.KeyCode = Keys.OemMinus And _
           Not e.KeyCode = Keys.OemPeriod Then
            e.SuppressKeyPress = True
        End If
        ' -------------------------------------------------------- no decimal point
        If e.KeyCode = Keys.OemPeriod Then
            e.SuppressKeyPress = True
        End If
        ' -------------------------------------------------------- no minus sign
        If e.KeyCode = Keys.OemMinus Then
            e.SuppressKeyPress = True
        End If
        ' ------------------------------------------------------- Handle pasted Text
        If e.Control AndAlso e.KeyCode = Keys.V Then
            ' --------------------------------------------------- Preview paste data (removing non-number characters)
            Dim pasteText As String = Clipboard.GetText
            Dim strippedText As String = ""
            Dim i As Integer = 0
            Do While i < pasteText.Length
                If Char.IsDigit(pasteText(i)) Then
                    strippedText = strippedText + pasteText(i).ToString
                End If
                i = (i + 1)
            Loop
            ' --------------------------------------------------- If there were non-numbers in the pasted text
            If strippedText <> pasteText Then

                e.SuppressKeyPress = True
                ' ----------------------------------------------- OPTIONAL: Manually insert text stripped of non-numbers
                Dim tbox As ComboBox = CType(sender, ComboBox)
                Dim start As Integer = tbox.SelectionStart
                Dim newTxt As String = tbox.Text
                newTxt = newTxt.Remove(tbox.SelectionStart, tbox.SelectionLength)
                ' ----------------------------------------------- remove highlighted text
                newTxt = newTxt.Insert(tbox.SelectionStart, strippedText)
                ' ----------------------------------------------- paste
                tbox.Text = newTxt
                tbox.SelectionStart = start + strippedText.Length
            Else
                e.SuppressKeyPress = False
            End If
        End If
    End Sub

    ' =============================================================
    '    SmootValue Adaptive filter - Double version
    ' =============================================================
    Friend Sub SmoothValue_Pow_Adaptive(ByRef value As Double, ByVal new_value As Double, ByVal speed As Double)
        Dim delta As Double = new_value - value
        Dim delta_sgn As Double = Math.Sign(delta)
        Dim delta_abs As Double = Math.Abs(delta)
        Dim delta_pow As Double = (delta_abs ^ 2) * speed
        If value <> 0 Then delta_pow /= Math.Abs(value)
        If delta_pow >= delta_abs Then
            value = new_value
        Else
            value += delta_sgn * delta_pow
        End If
        ' ------------------ if not equal to itself then it is a NAN
        If value <> value Then value = 0
    End Sub


    Friend Function SingleToInt32(ByVal n As Single) As Int32
        If n < Int32.MinValue Or n > Int32.MaxValue Then Return 0
        Return CInt(n)
    End Function
    Friend Function StringToInt32(ByVal s As String) As Int32
        Dim n As Double = Val(s)
        If n < Int32.MinValue Or n > Int32.MaxValue Then Return 0
        Return CInt(n)
    End Function

    Friend Function StringToUInt16(ByVal s As String) As UInt16
        Dim n As Double = Val(s)
        If n < UInt16.MinValue Or n > UInt16.MaxValue Then Return 0
        Return CUShort(n)
    End Function

    Friend Function StringToByte(ByVal s As String) As Byte
        Dim n As Double = Val(s)
        If n < 0 Or n > 255 Then Return 0
        Return CByte(n)
    End Function


End Module
